using System;
using System.Collections.Generic;
using System.Text;

namespace GeoMapLib
{
     public class clsDatum
    {

        public class clsEllipsoid
        {
            public double a;
            public double b;
            public double f;
            public double e;
            public double e2;

            public clsEllipsoid(double a, double b, double f, double e, double e2)
            {
                this.a = a;
                this.b = b;
                this.f = f;
                this.e = e;
                this.e2 = e2;
            }
        }

        public class clsConvFactor
        {
            public double t1;
            public double t2;
            public double t3;

            public clsConvFactor(double t1, double t2, double t3)
            {
                this.t1 = t1;
                this.t2 = t2;
                this.t3 = t3;
            }
        }

        public class clsXYZ
        {
            public double x;
            public double y;
            public double z;
            public int err;
        }

        public class clsLatlon
        {
            public double lat;
            public double lon;
            public double height;
            public int err;
        }


        public static clsEllipsoid BESSEL = new clsEllipsoid(6377397.155, 6356078.96283244, 1 / 299.152813, 0.0816968311952638, 0.00667437222734743);


        public static clsEllipsoid GRS80 = new clsEllipsoid(6378137, 6356752.31414036, 1 / 298.257222101, 0.0818191910428158, 0.00669438002290079);

        static clsConvFactor PAR3_ITRF94_TOKYO97 = new clsConvFactor(146.414, -507.337, -680.507);

        const int ITRF94 = 1;
        const int TOKYO97 = 2;

        const double DEG2RAD = 0.01745329251994;
        const double RAD2DEG = 57.295779513082;


        //*******************************************************************
        //
        // Enn̈ܓxox{nnɕϊB
        //
        // @\FEnniITRF94-GRS80)ł̈ܓxox{nn̈ܓxox
        //       ɕϊB
        //
        // F
        //   (in)
        //		convMode: ϊ@
        //                  1: S𓯂p[^ōs(n`̘clȂjB
        //                        ʂTokyo97-BESSEL
        //                  2: n斈̃p[^ōsBij
        //                        ʂTokyo-BESSEL
        //		lat: 	ܓxixPʁj
        //		lon:  	oxixPʁj
        //		height:	ȉ~̍im)
        //
        //   (out)
        //		newLat:    ܓxixPʁj
        //		newLon:    oxixPʁj
        //		newHeight: ȉ~̍im)
        //      ߂lF   G[R[h(0:I@:G[j
        //
        //------------------------------------------------------------------
        //
         public static int world2Tokyo(int convMode, double lat, double lon, double height,
                                     out double newLat, out double newLon, out double newHeight)
         {
             int nret;
             clsXYZ xyz = new clsXYZ();
             clsLatlon latlon = default(clsLatlon);
             double x = 0;
             double y = 0;
             double z = 0;

             xyz = latLon2XYZ(ITRF94, lat, lon, height);

             x = xyz.x + PAR3_ITRF94_TOKYO97.t1;
             y = xyz.y + PAR3_ITRF94_TOKYO97.t2;
             z = xyz.z + PAR3_ITRF94_TOKYO97.t3;

             latlon = xyz2LatLon(TOKYO97, x, y, z);

             newLat = latlon.lat;
             newLon = latlon.lon;
             newHeight = latlon.height;

             if ((convMode != 1)) nret = -1; else nret = 0;

             return nret;
         }


        //*******************************************************************
        //
        // {nn̈ܓxox𐢊EnnɕϊB
        //
        // @\F{nnł̈ܓxox𐢊EnniITRF94-GRS80)̈ܓxox
        //       ɕϊB
        //
        // F
        //   (in)
        //		convMode: ϊ@
        //                  1: S𓯂p[^ōs(n`̘clȂjB
        //                        ̍WTokyo97-BESSEL
        //                  2: n斈̃p[^ōsBij
        //                        ̍WTokyo-BESSEL
        //		lat: 	ܓxixPʁj
        //		lon:  	oxixPʁj
        //		height:	ȉ~̍im)
        //
        //   (out)
        //		newLat:    ܓxixPʁj
        //		newLon:    oxixPʁj
        //		newHeight: ȉ~̍im)
        //      ߂lF   G[R[h(0:I@:G[j
        //
        //------------------------------------------------------------------
        //
         public static int tokyo2World(int convMode, double lat, double lon, double height,
                                     out double newLat, out double newLon, out double newHeight)
        {
            int nret;
            clsXYZ xyz = new clsXYZ();
            clsLatlon latlon = default(clsLatlon);
            double x = 0;
            double y = 0;
            double z = 0;

            xyz = latLon2XYZ(TOKYO97, lat, lon, height);

            x = xyz.x - PAR3_ITRF94_TOKYO97.t1;
            y = xyz.y - PAR3_ITRF94_TOKYO97.t2;
            z = xyz.z - PAR3_ITRF94_TOKYO97.t3;

            latlon = xyz2LatLon(ITRF94, x, y, z);

            newLat = latlon.lat;
            newLon = latlon.lon;
            newHeight = latlon.height;

            if ((convMode != 1)) nret = -1; else nret = 0;

            return nret;
        }

        //*******************************************************************
        //
        // ܓxoxRWlɕϊB
        //
        // @\F̑nnɂāAnIW(ȉ~̖ʏ̍Wj
        //       RWɕϊB
        //
        // F
        //   (in)
        //		dataum:	nni 1:Enn(ITRF94-GRS80) 
        //                       2:{nniTOKYO97-BESSEL) )
        //		lat: 	ܓxixPʁj
        //		lon:  	oxixPʁj
        //		height:	ȉ~̍im)
        //
        // ߂l : XYZWl(IuWFNg)
        //          .x : XWim)
        //          .y : YWim)
        //          .z : ZWim)
        //          .err : G[R[h(0:I@:G[j
        //
        //------------------------------------------------------------------
        //
        public static clsXYZ latLon2XYZ(int datum, double lat, double lon, double height)
        {

            clsXYZ xyz = new clsXYZ();
            double a = 0;
            double e2 = 0;

            switch ((datum))
            {
                case ITRF94:
                    a = GRS80.a;
                    e2 = GRS80.e2;

                    break;
                case TOKYO97:
                    a = BESSEL.a;
                    e2 = BESSEL.e2;

                    break;
                default:
                    xyz.err = -1;
                    // o^ĂȂnn
                    return xyz;
            }

            double B = lat * DEG2RAD;
            double L = lon * DEG2RAD;

            double sinB = Math.Sin(B);
            double cosB = Math.Cos(B);
            double sinL = Math.Sin(L);
            double cosL = Math.Cos(L);

            double N = a / Math.Sqrt(1 - e2 * sinB * sinB);

            xyz.x = (N + height) * cosB * cosL;
            xyz.y = (N + height) * cosB * sinL;
            xyz.z = (N * (1 - e2) + height) * sinB;
            xyz.err = 0;

            return xyz;
        }



        //*******************************************************************
        //
        // RWlܓxoxɕϊB
        //
        // @\F̑nnɂāARW
        //       nIW(ȉ~̖ʏ̍WjɕϊB
        //
        // F
        //   (in)
        //		dataum:	nni 1:Enn(ITRF94-GRS80) 2:{nniTOKYO97-BESSEL) )
        //      x : XWim)
        //      y : YWim)
        //      z : ZWim)
        //
        // ߂l : ܓxox(IuWFNg)
        //		.lat: 	ܓxixPʁj
        //		.lon:  	oxixPʁj
        //		.height:	ȉ~̍im)
        //      .err : G[R[h(0:I@:G[j
        //
        //------------------------------------------------------------------
        //
        public static clsLatlon xyz2LatLon(int datum, double x, double y, double z)
        {

            clsLatlon latlon = new clsLatlon();
            double a, e2, B, L, f;

            switch ((datum))
            {
                case ITRF94:
                    a = GRS80.a;
                    e2 = GRS80.e2;
                    f = GRS80.f;

                    break;
                case TOKYO97:
                    a = BESSEL.a;
                    e2 = BESSEL.e2;
                    f = BESSEL.f;

                    break;
                default:
                    latlon.err = -1;
                    // o^ĂȂnn
                    return latlon;
            }

            L = Math.Atan2(y, x);

            double p = Math.Sqrt(x * x + y * y);
            double r = Math.Sqrt(p * p + z * z);

            double mu = Math.Atan2(z * (1 - f + e2 * a / r), p);
            double sinMu = Math.Sin(mu);
            double cosMu = Math.Cos(mu);
            double sinMu3 = sinMu * sinMu * sinMu;
            double cosMu3 = cosMu * cosMu * cosMu;

            double b1 = z * (1 - f) + e2 * a * sinMu3;
            double b2 = (1 - f) * (p - e2 * a * cosMu3);

            B = Math.Atan2(b1, b2);

            double sinB = Math.Sin(B);
            double cosB = Math.Cos(B);
            double height = p * cosB + z * sinB - a * Math.Sqrt(1 - e2 * sinB * sinB);

            latlon.lat = B * RAD2DEG;
            latlon.lon = L * RAD2DEG;
            latlon.height = height;
            latlon.err = 0;

            return latlon;
        }

    }

}
