using System;
using System.Collections.Generic;
using System.Text;

namespace GeoMapLib
{

    //***************************************************************************
    //                           @@@@ȉ~
    //---------------------------------------------------------------------------

    public class clsEllipsoid
    {
        public double m_a;   // a
        public double m_b;   // Za
        public double m_f;   // G
        public double m_e;   // ꗣS
        public double m_e2;  // êQ

        double m_CoeffA;     // qߐvZp
        double m_CoeffB;
        double m_CoeffC;
        double m_CoeffD;
        double m_CoeffE;
        double m_CoeffF;

        public void setParam(double a, double b)
        {
            try
            {
                double e2, e4, e6, e8, e10;
                m_a = a;
                m_b = b;
                m_f = (a - b) / a;
                m_e = Math.Sqrt((a * a - b * b) / (a * a));
                m_e2 = m_e * m_e;

                e2 = m_e2;
                e4 = e2 * e2;
                e6 = e4 * e2;
                e8 = e6 * e2;
                e10 = e8 * e2;

                m_CoeffA = 1.0 + 3.0 / 4.0 * e2 + 45.0 / 64.0 * e4 + 175.0 / 256.0 * e6 + 11025.0 / 16384.0 * e8 + 43659.0 / 65536.0 * e10;
                m_CoeffB = 3.0 / 4.0 * e2 + 15.0 / 16.0 * e4 + 525.0 / 512.0 * e6 + 2205.0 / 2047.0 * e8 + 72765.0 / 65536.0 * e10;
                m_CoeffC = 15.0 / 64.0 * e4 + 105.0 / 256.0 * e6 + 2205.0 / 4096.0 * e8 + 10395.0 / 16384.0 * e10;
                m_CoeffD = 35.0 / 512.0 * e6 + 315.0 / 2048 * e8 + 31185.0 / 131072.0 * e10;
                m_CoeffE = 315.0 / 16384.0 * e8 + 3465.0 / 65535.0 * e10;

                m_CoeffF = 693.0 / 131072.0 * e10;
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
            }

        }

        public clsEllipsoid(int ellipsoidNum)
        {
            switch (ellipsoidNum)
            {
                case Const.BESSEL:
                    setParam(6377397.155, 6356078.96283244);
                    break;

                case Const.GRS80:
                default:
                    setParam(6378137, 6356752.31414036);
                    break;
            }
        }

        // w肵ܓxł̎qߐ̋ȗa(M)߂B
        //
        // lat: ܓxiWA)
        //
        public double getMeridanRadius(double lat)
        {
            double s = Math.Sin(lat);
            double w = Math.Sqrt(1 - m_e2 * s * s);
            double m = m_a * (1 - m_e2) / (w * w * w);
            return m;
        }


        // w肵ܓxł̉Kѐ̋ȗa(N)߂B
        //
        // lat: ܓxiWA)
        //
        public double getPrimeVerticalRadius(double lat)
        {
            double s = Math.Sin(lat);
            double w = Math.Sqrt(1 - m_e2 * s * s);
            double n = m_a / w;
            return n;
        }


        // Q_Ԃ̋߂B
        public double getDistance(double lat1, double lon1, double lat2, double lon2)
        {
            double aveLat = (lat1 + lat2) / 2 * Const.DEG2RAD;

            double m = getMeridanRadius(aveLat);
            double n = getPrimeVerticalRadius(aveLat);

            double latDist = ((lat1 - lat2) * Const.DEG2RAD) * m;
            double lonDist = ((lon1 - lon2) * Const.DEG2RAD) * Math.Cos(aveLat) * n;

            double distance = Math.Sqrt(latDist * latDist + lonDist * lonDist);
            return distance;
        }

        // _̋߂
        //
        // points: xɌoxAyɈܓxxPʂŊi[Ă̂ƂB
        // 

        public double getDistance(stXY[] points, int numPoints)
        {
            try
            {
                double lon1, lat1, lon2, lat2, aveLat, m, n, latDist, lonDist, distance;
                int imax;

                if (numPoints < 2) return 0;

                lon1 = points[0].x * Const.DEG2RAD;
                lat1 = points[0].y * Const.DEG2RAD;
                distance = 0;
                imax = numPoints - 1;
                for (int i = 0; i < imax; i++)
                {
                    lon2 = points[i + 1].x * Const.DEG2RAD;
                    lat2 = points[i + 1].y * Const.DEG2RAD;

                    aveLat = (lat1 + lat2) / 2;

                    m = getMeridanRadius(aveLat);
                    n = getPrimeVerticalRadius(aveLat);

                    latDist = (lat1 - lat2) * m;
                    lonDist = (lon1 - lon2) * Math.Cos(aveLat) * n;

                    distance += Math.Sqrt(latDist * latDist + lonDist * lonDist);

                    lat1 = lat2;
                    lon1 = lon2;
                }

                return distance;

            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return 0;
            }
        }


        // ܓxphi1ܓxphi2܂ł̎qߐʒ(m)߂B
        //
        // phi1,phi2 : ܓxiWAj
        // 
        //QlunwvP24
        //
        public double getMeridianLength(double phi1, double phi2)
        {
            try
            {
                double a = m_CoeffA * (phi2 - phi1);
                double b = m_CoeffB * (Math.Sin(2 * phi2) - Math.Sin(2 * phi1)) / 2.0;
                double c = m_CoeffC * (Math.Sin(4 * phi2) - Math.Sin(4 * phi1)) / 4.0;
                double d = m_CoeffD * (Math.Sin(6 * phi2) - Math.Sin(6 * phi1)) / 6.0;
                double e = m_CoeffE * (Math.Sin(8 * phi2) - Math.Sin(8 * phi1)) / 8.0;
                double f = m_CoeffF * (Math.Sin(10 * phi2) - Math.Sin(10 * phi1)) / 10;

                return m_a * (1 - m_e2) * (a - b + c - d + e - f);
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return -1;
            }
        }


    }

}
