using System;
using System.Collections.Generic;
using System.Text;

namespace GeoMapLib
{
    public class clsGaussKruger
    {

        //***************************************************************************************
        //
        //                @KEXN[QeiJg[ejNX
        // 
        //***************************************************************************************


        public const double PI = Const.PI;

        public const double DEG2RAD = PI / 180;
        public const double RAD2DEG = 180 / PI;

        double m_radius;
        double m_e2;
        double m_A;
        double m_B;
        double m_C;
        double m_D;
        double m_E;
        double m_F;

        //
        //ellipsoidNum = ȉ~̔ԍ@0:GRS80  1:xbZ
        public clsGaussKruger(int ellipsoidNum)
        {
            double e2, e4 , e6, e8 , e10 ;
            try
            {
                clsEllipsoid ellipsoid = new clsEllipsoid(ellipsoidNum);
                m_radius = ellipsoid.m_a;
                m_e2 = ellipsoid.m_e2;

                e2 = m_e2;
                e4 = e2 * e2;
                e6 = e4 * e2;
                e8 = e6 * e2;
                e10 = e8 * e2;

                m_A = 1.0 + 3.0 / 4.0 * e2 + 45.0 / 64.0 * e4 + 175.0 / 256.0 * e6 + 11025.0 / 16384.0 * e8 + 43659.0 / 65536.0 * e10;
                m_B = 3.0 / 4.0 * e2 + 15.0 / 16.0 * e4 + 525.0 / 512.0 * e6 + 2205.0 / 2047.0 * e8 + 72765.0 / 65536.0 * e10;
                m_C = 15.0 / 64.0 * e4 + 105.0 / 256.0 * e6 + 2205.0 / 4096.0 * e8 + 10395.0 / 16384.0 * e10;
                m_D = 35.0 / 512.0 * e6 + 315.0 / 2048.0 * e8 + 31185.0 / 131072.0 * e10;
                m_E = 315.0 / 16384.0 * e8 + 3465.0 / 65535.0 * e10;

                m_F = 693.0 / 131072.0 * e10;
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
            }
        }


        // ܓxphi1ܓxphi2܂ł̎qߐʒ(m)߂B
        // 
        //QlunwvP24
        //
        public double MeridianLength(double phi1, double phi2)
        {
            try
            {
                double a = 0;
                double b = 0;
                double c = 0;
                double d = 0;
                double e = 0;
                double f = 0;

                a = m_A * (phi2 - phi1);
                b = m_B * (Math.Sin(2 * phi2) - Math.Sin(2 * phi1)) / 2.0;
                c = m_C * (Math.Sin(4 * phi2) - Math.Sin(4 * phi1)) / 4.0;
                d = m_D * (Math.Sin(6 * phi2) - Math.Sin(6 * phi1)) / 6.0;
                e = m_E * (Math.Sin(8 * phi2) - Math.Sin(8 * phi1)) / 8.0;
                f = m_F * (Math.Sin(10 * phi2) - Math.Sin(10 * phi1)) / 10;

                return m_radius * (1 - m_e2) * (a - b + c - d + e - f);
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return -1;
            }
        }


        // ܓxoxXYWɕϊB
        //
        //  (in)
        //	lat0	F_̈ܓx
        //	lon0	F_ioǰox
        //	m0	    FkڌW
        //	lat     : ܓxixj
        //   lon     : oxixj
        //
        //  (out)
        //    x : xWijikj
        //@@y : y@@@@@ij
        //
        //	QlunwvP204
        //
        public int latlon2xy(double lat0, double lon0, double m0, double lat, double lon, out double x, out double y)
        {


double phi,lamda,phi0,lamda0,B,s,c,c2,c4,t,t2,t4,h2,h4;
double N, lamda2, lamdac2, lamdac4, term1, term2;

            x = 0.0;
            y = 0.0;
            try
            {

                phi = lat * DEG2RAD;
                lamda = lon * DEG2RAD;
                phi0 = lat0 * DEG2RAD;
                lamda0 = lon0 * DEG2RAD;

                lamda -= lamda0;

                s = Math.Sin(phi);

                c = Math.Cos(phi);
                c2 = c * c;
                c4 = c2 * c2;

                t = Math.Tan(phi);
                t2 = t * t;
                t4 = t2 * t2;

                h2 = m_e2 * c2 / (1 - m_e2);
                h4 = h2 * h2;

                N = m_radius / Math.Sqrt(1 - m_e2 * s * s);

                lamda2 = lamda * lamda;
                lamdac2 = lamda2 * c2;
                lamdac4 = lamdac2 * lamdac2;

                B = MeridianLength(phi0, phi);

                term1 = lamdac2 * (5 - t2 + 9 * h2 + 4 * h4) / 24.0;
                term2 = lamdac4 * (61 - 58 * t2 + t4 + 270 * h2 - 330 * h2 * t2) / 720.0;
                x = m0 * (B + N * s * c * lamda2 * (0.5 + term1 + term2));

                term1 = lamdac2 * (1 - t2 + h2) / 6.0;
                term2 = lamdac4 * (5 - 18 * t2 + t4 + 14 * h2 - 58 * h2 * t2) / 120.0;
                y = m0 * N * c * lamda * (1 + term1 + term2);

                return 0;
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return -1;
            }
        }



        //XYWܓxoxɕϊB
        //
        //  (in)
        //	lat0	F_̈ܓx
        //	lon0	F_ioǰox
        //	m0	    FkڌW
        //    x      : xWijikj
        //@@y      : y@@@@@ij
        //
        //  (out)
        //	lat     : ܓxixj
        //   lon     : oxixj
        //
        //	QlunwvP208
        //

        public int xy2latlon(double lat0, double lon0, double m0, double x, double y, out double lat, out double lon)
        {


            double phi0, lamda0, meridian0, phi1, meridian, s, c, deltaPhi, sq;
            double N1, R1, eta2, t, t2, t4, yN1, yN2, yN4, term1, term2, phi, lamda;

            lat = 0.0;
            lon = 0.0;

            try
            {

                phi0 = lat0 * DEG2RAD;
                lamda0 = lon0 * DEG2RAD;

                meridian0 = x / m0 + MeridianLength(0.0, phi0);
                y = y / m0;

                phi1 = phi0;
                while ((true))
                {
                    meridian = MeridianLength(0.0, phi1);
                    s = Math.Sin(phi1);
                    deltaPhi = (meridian0 - meridian) * Math.Pow(1 - m_e2 * s * s, 1.5) / (m_radius * (1 - m_e2));
                    if ((Math.Abs(deltaPhi) < 1E-10)) break; // TODO: might not be correct. Was : Exit While

                    phi1 += deltaPhi;
                }

                s = Math.Sin(phi1);
                c = Math.Cos(phi1);
                sq = Math.Sqrt(1 - m_e2 * s * s);
                N1 = sq / m_radius;
                R1 = (m_radius * (1 - m_e2)) / (sq * sq * sq);
                eta2 = (c * c) * m_e2 / (1 - m_e2);
                t = Math.Tan(phi1);
                t2 = t * t;
                t4 = t2 * t2;

                yN1 = y * N1;
                yN2 = yN1 * yN1;
                yN4 = yN2 * yN2;

                term1 = -yN2 * (5 + 3 * t2 + eta2 - 9 * t2 * eta2) / 12.0;
                term2 = yN4 * (61 + 90 * t2 + 45 * t4) / 360.0;
                phi = phi1 - y * yN1 * t * (1 + term1 + term2) / 2.0 / R1;

                term1 = -yN2 * (1 + 2 * t2 + eta2) / 6.0;
                term2 = yN4 * (5 + 28 * t2 + 24 * t4) / 120.0;
                lamda = lamda0 + yN1 / c * (1 + term1 + term2);

                lat = phi * RAD2DEG;
                lon = lamda * RAD2DEG;

                return 0;
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return -1;
            }
        }


    }

}
