﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Data;
using System.Diagnostics;
using System.Windows.Forms;
using GeoMapLib;

namespace GeoMapLib
{

    public class  clsHeimenChokkaku2
    {
        //***************************************************************************************
        //
        //                　　　　　　　平面直角座標系クラス
        // 
        //***************************************************************************************
        
        //
        // 計算式
        //　「Gauss-Kruber投影における経緯度座標及び平面直角座標相互間の
        //    座標換算についてのより簡明な計算方法」
        //     国土地理院　企画部　河瀬和重
        //

        double sinh(double x) { return 0.5f * (Math.Exp(x) - Math.Exp(-x)); }
        double cosh(double x) { return 0.5f * (Math.Exp(x) + Math.Exp(-x)); }
        double arctanh(double x) { return 0.5f * Math.Log((1 + x) / (1 - x)); }

        const int jt = 5;
        const int jt2 = 2 * jt;
        const double a = 6378137;
        const double rf = 298.257222101;
        const double m0 = 0.9999;
        const double s2r = Math.PI / 648000;
        const double n = 0.5 / (rf - 0.5);
        const double n15 = 1.5 * n;
        const double anh = 0.5f * a / (1 + n);
        int m_keiNum;

        public clsHeimenChokkaku2(int keiNum)
        {
            m_keiNum = keiNum;
        }

        // 該当緯度の2 倍角の入力により赤道からの子午線弧長を求める関数
        double Merid(double ep, double[] e, double phi2)
        {
            double dc = 2.0f * Math.Cos(phi2);
            double[] s = new double[jt2 + 2];
            double[] t = new double[jt2 + 1];
            s[1] = Math.Sin(phi2);
            for (int i = 1; i <= jt2; i++)
            {
                s[i + 1] = dc * s[i] - s[i - 1];
                t[i] = (1.0f / i - 4.0f * i) * s[i];
            }
            double sum = 0.0;
            double c1 = ep;
            double c2 = 0;
            int j = jt;
            while (j > 0)
            {
                c2 = phi2;
                double c3 = 2.0;
                int l = j;
                int m = 0;
                while (l > 0)
                {
                    c2 += (c3 /= e[l--]) * t[++m] + (c3 *= e[2 * j - l]) * t[++m];
                }
                sum += c1 * c1 * c2; c1 /= e[j--];
            }
            return anh * (sum + phi2);
        }

        public int latlon2xy( double lat, double lon, out double x, out double y )
        {
            x = 0;
            y = 0;
            double gmm = 0;
            double m = 1;
            int num = m_keiNum;
            try
            {
                double phirad = lat * 3600 * s2r;
                double lmbdsec = lon * 3600;

                double nsq = n * n;
                double e2n = 2 * Math.Sqrt(n) / (1 + n);
                double ra = 2 * anh * m0 * (1 + nsq / 4 + nsq * nsq / 64);
                double[] e = new double[jt * 2 + 1];
                double ep = 1.0;
                double[] alp = new double[6];
                for (int k = 1; k <= jt; k++)
                {
                    ep *= e[k] = n15 / k - n;
                    e[k + jt] = n15 / (k + jt) - n;
                }

                // 展開パラメータの事前入力
                alp[1] = (1.0 / 2 + (-2.0 / 3 + (5.0 / 16 + (41.0 / 180 - 127.0 / 288 * n) * n) * n) * n) * n;
                alp[2] = (13.0 / 48 + (-3.0 / 5 + (557.0 / 1440 + 281.0 / 630 * n) * n) * n) * nsq;
                alp[3] = (61.0 / 240 + (-103.0 / 140 + 15061.0 / 26880 * n) * n) * n * nsq;
                alp[4] = (49561.0 / 161280 - 179.0 / 168 * n) * nsq * nsq;
                alp[5] = 34729.0 / 80640 * n * nsq * nsq;

                // 平面直角座標の座標系原点の緯度を度単位で、経度を分単位で格納
                double[] phi0 = new double[] { 0, 33, 33, 36, 33, 36, 36, 36, 36, 36, 40, 44, 44, 44, 26, 26, 26, 26, 20, 26 };
                double[] lmbd0 = new double[] { 0, 7770, 7860, 7930, 8010, 8060, 8160, 8230, 8310, 8390, 8450, 8415, 8535, 8655, 8520, 7650, 7440, 7860, 8160, 9240 };


                // 実際の計算実行部分
                double sphi = Math.Sin(phirad);
                double nphi = (1 - n) / (1 + n) * Math.Tan(phirad);
                double dlmbd=( lmbdsec-lmbd0[num]*60)*s2r;
                double sdlmbd = Math.Sin(dlmbd);
                double cdlmbd = Math.Cos(dlmbd);
                double tchi = sinh(arctanh(sphi) - e2n * arctanh(e2n * sphi));
                double cchi = Math.Sqrt(1 + tchi * tchi);
                double xi = Math.Atan2(tchi, cdlmbd);
                double xip = xi;
                double eta = arctanh(sdlmbd / cchi);
                double etap = eta;
                double sgm = 1;
                double tau = 0;
                for (int j = alp.Length - 1; j > 0; j--)
                {
                    double alsin = alp[j] * Math.Sin(2 * j * xip);
                    double alcos = alp[j] * Math.Cos(2 * j * xip);
                    xi += alsin * cosh(2 * j * etap);
                    eta += alcos * sinh(2 * j * etap);
                    sgm += 2 * j * alcos * cosh(2 * j * etap);
                    tau += 2 * j * alsin * sinh(2 * j * etap);
                }
                x = ra * xi - m0 * Merid(ep, e, 2 * phi0[num] * 3600 * s2r);
                y = ra * eta;
                gmm = Math.Atan2(tau * cchi * cdlmbd + sgm * tchi * sdlmbd,
                                        sgm * cchi * cdlmbd - tau * tchi * sdlmbd);
                m = ra / a * Math.Sqrt(
                    (sgm * sgm + tau * tau) / (tchi * tchi + cdlmbd * cdlmbd) * (1 + nphi * nphi));

                return 0;
            }
            catch (Exception ex)
            {
                return -100;
            }
        }

        public int xy2latlon( double x, double y, out double phi, out double lmbd)
        {
            phi = 0;
            lmbd = 0;
            double gmm = 0;
            double m = 1;
            int num = m_keiNum;
            try
            {
                double nsq = n * n;
                double e2n = 2 * Math.Sqrt(n) / (1 + n);
                double ra = 2 * anh * m0 * (1 + nsq / 4 + nsq * nsq / 64);
                double[] e = new double[jt * 2 + 1];
                double ep = 1.0;
                double[] alp = new double[6];
                for (int k = 1; k <= jt; k++)
                {
                    ep *= e[k] = n15 / k - n;
                    e[k + jt] = n15 / (k + jt) - n;
                }
                double[] beta = new double[6];
                double[] dlt = new double[7];

                // 展開パラメータの事前入力
                beta[1] = (1.0 / 2 + (-2.0 / 3 + (37.0 / 96 + (-1.0 / 360 - 81.0 / 512 * n) * n) * n) * n) * n;
                beta[2] = (1.0 / 48 + (1.0 / 15 + (-437.0 / 1440 + 46.0 / 105 * n) * n) * n) * nsq;
                beta[3] = (17.0 / 480 + (-37.0 / 840 - 209.0 / 4480 * n) * n) * n * nsq;
                beta[4] = (4397.0 / 161280 - 11.0 / 504 * n) * nsq * nsq;
                beta[5] = 4583.0 / 161280 * n * nsq * nsq;
                dlt[1] = (2 + (-2.0 / 3 + (-2.0 + (116.0 / 45 + (26.0 / 45 - 2854.0 / 675 * n) * n) * n) * n) * n) * n;
                dlt[2] = (7.0 / 3 + (-8.0 / 5 + (-227.0 / 45 + (2704.0 / 315 + 2323.0 / 945 * n) * n) * n) * n) * nsq;
                dlt[3] = (56.0 / 15.0 + (-136.0 / 35 + (-1262.0 / 105 + 73814.0 / 2835 * n) * n) * n) * n * nsq;
                dlt[4] = (4279.0 / 630 + (-332.0 / 35 - 399572.0 / 14175 * n) * n) * nsq * nsq;
                dlt[5] = (4174.0 / 315 - 144838.0 / 6237 * n) * n * nsq * nsq;
                dlt[6] = 601676.0 / 22275 * nsq * nsq * nsq;

                // 平面直角座標の座標系原点の緯度を度単位で、経度を分単位で格納
                double[] phi0 = new double[] { 0, 33, 33, 36, 33, 36, 36, 36, 36, 36, 40, 44, 44, 44, 26, 26, 26, 26, 20, 26 };
                double[] lmbd0 = new double[] { 0, 7770, 7860, 7930, 8010, 8060, 8160, 8230, 8310, 8390, 8450, 8415, 8535, 8655, 8520, 7650, 7440, 7860, 8160, 9240 };


                // 実際の計算実行部分
                double xi = (x + m0 * Merid(ep, e, 2 * phi0[num] * 3600 * s2r)) / ra;
                double xip = xi;
                double eta = y / ra;
                double etap = eta;
                double sgmp = 1;
                double taup = 0;
                for (int j = beta.Length - 1; j > 0; j--)
                {
                    double besin = beta[j] * Math.Sin(2 * j * xi);
                    double becos = beta[j] * Math.Cos(2 * j * xi);
                    xip -= besin * cosh(2 * j * eta);
                    etap -= becos * sinh(2 * j * eta);
                    sgmp -= 2 * j * becos * cosh(2 * j * eta);
                    taup += 2 * j * besin * sinh(2 * j * eta);
                }

                double sxip = Math.Sin(xip);
                double cxip = Math.Cos(xip);
                double shetap = sinh(etap);
                double chetap = cosh(etap);
                double chi = Math.Asin(sxip / chetap);
                phi = chi;
                for (int j = dlt.Length - 1; j > 0; j--)
                {
                    phi += dlt[j] * Math.Sin(2 * j * chi);
                }
                double nphi = (1 - n) / (1 + n) * Math.Tan(phi);

                lmbd = lmbd0[num] * 60 + Math.Atan2(shetap, cxip) / s2r;    // sec
                lmbd /= 3600;   // degree
                phi = phi / s2r / 3600; 

                gmm = Math.Atan2(taup * cxip * chetap + sgmp * sxip * shetap,
                                        sgmp * cxip * chetap - taup * sxip * shetap); // radian
                m = ra / a * Math.Sqrt(
                    (cxip * cxip + shetap * shetap) / (sgmp * sgmp + taup * taup) * (1 + nphi * nphi));
                return 0;
            }
            catch (Exception ex)
            {
                return -100;
            }
        }
    }

    public class clsHeimenChokkaku
    {

        //***************************************************************************************
        //
        //                　　　　　　　平面直角座標系クラス
        // 
        //        こちらの実装は緯度方向に300ｍ程度違っている
        // 　　
        //***************************************************************************************

        public double m_lat0;
        public double m_lon0;
        public double m_m0 = 0.9999;

        public clsGaussKruger m_gaussKruger;

        public clsHeimenChokkaku(int keiNum)
        {

            try
            {
                switch (keiNum)
                {
                    case 1:
                        m_lon0 = 129.5;
                        m_lat0 = 33;
                        break;
                    case 2:
                        m_lon0 = 131;
                        m_lat0 = 33;
                        break;
                    case 3:
                        m_lon0 = 132 + 10.0 / 60;
                        m_lat0 = 36;
                        break;
                    case 4:
                        m_lon0 = 133.5;
                        m_lat0 = 33;
                        break;
                    case 5:
                        m_lon0 = 134 + 20.0 / 60;
                        m_lat0 = 36;
                        break;
                    case 6:
                        m_lon0 = 136;
                        m_lat0 = 36;
                        break;
                    case 7:
                        m_lon0 = 137 + 10.0 / 60;
                        m_lat0 = 36;
                        break;
                    case 8:
                        m_lon0 = 138.5;
                        m_lat0 = 36;
                        break;
                    case 9:
                        m_lon0 = 139 + 50.0 / 60;
                        m_lat0 = 36;
                        break;
                    case 10:
                        m_lon0 = 140 + 50.0 / 60;
                        m_lat0 = 40;
                        break;
                    case 11:
                        m_lon0 = 140 + 15.0 / 60;
                        m_lat0 = 44;
                        break;
                    case 12:
                        m_lon0 = 142 + 15.0 / 60;
                        m_lat0 = 44;
                        break;
                    case 13:
                        m_lon0 = 144 + 15.0 / 60;
                        m_lat0 = 44;
                        break;
                    case 14:
                        m_lon0 = 142;
                        m_lat0 = 26;
                        break;
                    case 15:
                        m_lon0 = 127.5;
                        m_lat0 = 26;
                        break;
                    case 16:
                        m_lon0 = 124;
                        m_lat0 = 26;
                        break;
                    case 17:
                        m_lon0 = 131;
                        m_lat0 = 26;
                        break;
                    case 18:
                        m_lon0 = 136;
                        m_lat0 = 20;
                        break;
                    case 19:
                        m_lon0 = 154;
                        m_lat0 = 26;
                        break;
                }

                m_gaussKruger = new clsGaussKruger(0);
                //0:GRS80
                if (m_gaussKruger == null)
                {
                    MessageBox.Show("平面直角座標系の初期化ができません。");
                    return;
                }

            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }

        }

        //緯度経度をXY座標に変換する。
        //
        //  (in)
        //	lat     : 緯度（度）
        //   lon     : 経度（度）
        //
        //  (out)
        //    x : x座標（ｍ）（南北方向）
        //　　y : y　　　　　（東西方向）

        public object latlon2xy(double lat, double lon, ref double x, ref double y)
        {
            try
            {
                int nret = 0;
                nret = m_gaussKruger.latlon2xy(m_lat0, m_lon0, m_m0, lat, lon, out x, out y);
                return nret;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
                return -1;
            }
        }

        //XY座標を緯度経度に変換する。
        //
        //  (in)
        //    x      : x座標（ｍ）（南北方向）
        //　　y      : y　　　　　（東西方向）
        //
        //  (out)
        //	lat     : 緯度（度）
        //   lon     : 経度（度）

        public object xy2latlon(double x, double y, ref double lat, ref double lon)
        {
            try
            {
                int nret = 0;
                nret = m_gaussKruger.xy2latlon(m_lat0, m_lon0, m_m0, x, y, out lat, out lon);
                return nret;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
                return -1;
            }
        }


        //投影原点の緯度経度を取得する。
        //
        public static object getOriginLatLon(int keiNum, ref double lat, ref double lon)
        {
            try
            {
                clsHeimenChokkaku meObj = new clsHeimenChokkaku(keiNum);
                lat = meObj.m_lat0;
                lon = meObj.m_lon0;
                return 0;

            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
                return -1;
            }

        }

    }

}
