using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using System.Drawing;
using System.Diagnostics;
using System.ComponentModel;
using System.Net;
using System.Web;
using Microsoft.Win32;


namespace GeoMapLib
{
    //***************************************************************************
    //                           @@@@萔
    //---------------------------------------------------------------------------

    public static class Const
    {
        public const double PI = 3.1415926535897932;

        public const double DEG2RAD = PI / 180.0;
        public const double RAD2DEG = 180.0 / PI;

        public const double DEG2MSEC = 3600 * 1000.0;
        public const double MSEC2DEG = 1.0 / 3600.0 / 1000.0;


        public const int GRS80 = 0;
        public const int BESSEL = 1;

        public const int DATUM_WORLD = 1;
        public const int DATUM_TOKYO = 2;

        public const long MAX_LONG = (1L << 63);

        public const double MAX_DOUBLE = double.MaxValue;

    }

    //***************************************************************************
    //                           @@@Wi[p\́ANX
    //---------------------------------------------------------------------------

    //_,xNgȂxAygŕ\\
    [Serializable()]
    public struct stXY
    {
        public stXY(double x, double y)
        {
            this.x = x;
            this.y = y;
        }

        public double x;
        public double y;

        // w肵_Ƃ̋ԂB
        public double distance(stXY point)
        {
            double xd = point.x - x;
            double yd = point.y - y;
            return Math.Sqrt(xd * xd + yd * yd);
        }
    }

    // `\\
    [Serializable()]
    public struct stGeoRect
    {
        public stGeoRect(double topLat, double leftLon, double bottomLat, double rightLon)
        {
            this.topLat = topLat;
            this.leftLon = leftLon;
            this.bottomLat = bottomLat;
            this.rightLon = rightLon;
        }

        public void set(double topLat, double leftLon, double bottomLat, double rightLon)
        {
            this.topLat = topLat;
            this.leftLon = leftLon;
            this.bottomLat = bottomLat;
            this.rightLon = rightLon;
        }

        public double topLat;
        public double leftLon;
        public double bottomLat;
        public double rightLon;
    }

    // nI͈͂\\
    [Serializable()]
    public struct stGeoMinMax
    {
        public stGeoMinMax(double minLat, double minLon, double maxLat, double maxLon)
        {
            this.minLat = minLat;
            this.minLon = minLon;
            this.maxLat = maxLat;
            this.maxLon = maxLon;
        }

        public void set(double minLat, double minLon, double maxLat, double maxLon)
        {
            this.minLat = minLat;
            this.minLon = minLon;
            this.maxLat = maxLat;
            this.maxLon = maxLon;
        }

        public void get(out double minLat, out double minLon, out double maxLat, out double maxLon)
        {
            minLat = this.minLat;
            minLon = this.minLon;
            maxLat = this.maxLat;
            maxLon = this.maxLon;
        }

        public double minLat;
        public double minLon;
        public double maxLat;
        public double maxLon;
    }


    //***************************************************************************
    //                          GPS̎Mf[^\\
    //---------------------------------------------------------------------------

    [Serializable()]
    public struct stGpsSatelliteInfo
    {
        public byte id;
        public byte signalLevel;     // Pʂ͎M@ɂB
        public short azimuth;         // degree
        public short elevation;       // degree
    }

    [Serializable()]
    public struct stGpsRecvInfo
    {
        public byte numSattelites;  // Młq
        public byte DGPS;           // DGPS␳^Cvi0:␳Ȃ 1:j
        public float GDOP;
        public float PDOP;
        public float HDOP;
    }

    //***************************************************************************
    //                      ̑NXA\́AfQ[g
    //***************************************************************************


    //***************************************************************************
    //                           @@nIȊ֐
    //***************************************************************************

    public static class GeoF
    {

        // xɕϊB
        //
        // nn͐EnnAߕ̓KEXN[Qeig嗦=1.0jōsB
        //
        // Pʂ͓xm
        //
        public static int length2Degree(double centerLat, double centerLon, double distance,
                                        out double latLength, out double lonLength)
        {
            latLength = 0.0;
            lonLength = 0.0;
            try
            {
                clsGaussKruger gk = new clsGaussKruger(Const.DATUM_WORLD);
                double lat, lon;
                gk.xy2latlon(centerLat, centerLon, 1.0, distance, 0, out lat, out lon); // k
                latLength = lat - centerLat;

                gk.xy2latlon(centerLat, centerLon, 1.0, 0, distance, out lat, out lon); // 
                lonLength = lon - centerLon;

                return 0;
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return -1;
            }
        }

        // xɕϊB
        //
        // nn͐EnnAߕ̓KEXN[Qeig嗦=1.0jōsB
        //
        // degree : ܓx͌ox̋Bܓx̏ꍇdegreeIsLatitudeTRUE,ox̏ꍇfalseɂB
        //
        // Pʂ͓xm
        //
        public static double degree2Length(double centerLat, double centerLon, double degree, bool degreeIsLatitude)
        {
            try
            {
                clsGaussKruger gk = new clsGaussKruger(Const.DATUM_WORLD);
                double x, y, distance;

                if (degreeIsLatitude)
                {
                    // k
                    gk.latlon2xy(centerLat, centerLon, 1.0, centerLat + degree, centerLon, out x, out y);
                    distance = x;
                }
                else
                {
                    // 
                    gk.latlon2xy(centerLat, centerLon, 1.0, centerLat, centerLon + degree, out x, out y);
                    distance = y;
                }

                return distance;
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return 0.0;
            }
        }

        // S_̊pxƋŎw肵_̈ʒu߂B
        //
        // nn͐EnnAߕ̓KEXN[Qeig嗦=1.0jōsB
        //
        // Pʂ͓xm

        public static int getPolarPoint(double centerLat, double centerLon, double degree, double mDistance,
                                out double lat, out double lon)
        {
            lat = 0;
            lon = 0;
            try
            {
                clsGaussKruger gk = new clsGaussKruger(Const.DATUM_WORLD);

                double radian = degree * Const.DEG2RAD;
                double y = mDistance * Math.Cos( radian );
                double x = mDistance * Math.Sin( radian );
                gk.xy2latlon( centerLat, centerLon,1.0, x, y, out lat, out lon );
                return 0;
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return -1;
            }
        }

        // S_̊pxƒ߂B
        //
        // nn͐EnnAߕ̓KEXN[Qeig嗦=1.0jōsB
        //
        // Pʂ͓x,m

        public static int getPolarAngle(double centerLat, double centerLon, double lat, double lon,
                               out double degree, out double mDistance )
        {
            degree= 0;
            mDistance = 0;
            try
            {
                clsGaussKruger gk = new clsGaussKruger(Const.DATUM_WORLD);

                double x, y;
                gk.latlon2xy(centerLat, centerLon, 1.0,lat,lon, out x,out y);
                degree = Math.Atan2(x, y) * Const.RAD2DEG;
                mDistance = Math.Sqrt(x * x + y * y);
                
                return 0;
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return -1;
            }
        }



        // _~ɑ݂邩肷B
        //
        // ~̃p[^setPointIsInsideCircle()Őݒ肵ĂB
        //
        static double m_centerLat;
        static double m_centerLon;
        static double m_radius;
        static double m_minLat, m_minLon, m_maxLat, m_maxLon;
        static double m_minLat2, m_minLon2, m_maxLat2, m_maxLon2;
        static clsEllipsoid m_ellipsoid;

        public static bool pointIsInsideCircle(double lat, double lon)
        {
            try
            {
                if (lat < m_minLat || lon < m_minLon || lat > m_maxLat || lon > m_maxLon) return false;
                if (lat >= m_minLat2 && lon >= m_minLon2 && lat <= m_maxLat2 && lon <= m_maxLon2) return true;

                double radius = m_ellipsoid.getDistance(m_centerLat, m_centerLon, lat, lon) / 1000.0;
                return (radius <= m_radius);
            }
            catch (Exception)
            {
                return false;
            }
        }

        public static int setPointIsInsideCircle(double centerLat, double centerLon, double radius)
        {
            try
            {
                m_ellipsoid = new clsEllipsoid(Const.GRS80);
                m_centerLat = centerLat;
                m_centerLon = centerLon;
                m_radius = radius;

                // Oڋ`
                double latLength, lonLength;
                length2Degree(m_centerLat, m_centerLon, m_radius * 1000.0, out latLength, out lonLength);
                m_minLat = m_centerLat - latLength;
                m_minLon = m_centerLon - lonLength;
                m_maxLat = m_centerLat + latLength;
                m_maxLon = m_centerLon + lonLength;

                // ڋ`
                length2Degree(m_centerLat, m_centerLon, m_radius * 1000.0 / 1.42, out latLength, out lonLength);
                m_minLat2 = m_centerLat - latLength;
                m_minLon2 = m_centerLon - lonLength;
                m_maxLat2 = m_centerLat + latLength;
                m_maxLon2 = m_centerLon + lonLength;

                return 0;
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return -1;
            }
        }
    }

    //***************************************************************************
    //                           @@rpNX
    //***************************************************************************

    // Ɋ܂܂鐔𐮐ƂĔrB
    //
    // ͑SpApǂłBrۂɑSp𔼊pɕϊĂB
    // 񂩂琔Ȃꍇ0ƂB
    // getIntegerRegexnull͍̎ŏɌgB
    //
    /*
    public class clsComparerStringWithInteger : IComparer
    {
        Regex m_getIntegerRegex;
        public clsComparerStringWithInteger(Regex getIntegerRegex)
        {
            if (getIntegerRegex == null) m_getIntegerRegex = new Regex(@"[0-9]+");
            else m_getIntegerRegex = getIntegerRegex;
        }

        int IComparer.Compare(Object x, Object y)
        {
            if (x == null || y == null) return 0;

            string sx = Util.getHankakuString(x.ToString());
            string sy = Util.getHankakuString(y.ToString());
            int ix = 0, iy = 0;
            Match mx = m_getIntegerRegex.Match(sx);
            if (mx != Match.Empty) int.TryParse(mx.ToString(), out ix);
            Match my = m_getIntegerRegex.Match(sy);
            if (my != Match.Empty) int.TryParse(my.ToString(), out iy);

            return (ix - iy);
        }
    }
    */



    //***************************************************************************
    //                           @̑([eBeBj
    //***************************************************************************


    public static class Util
    {
        private static string m_lastAccessDirectory;
        private static Control m_waitMessageControl;
        private static FormWaitMessage m_waitMessageForm;
        private static int m_httpTimeoutMilliSec = 100000; // 100b

        //***************************************************************************
        //                           @@@@z
        //---------------------------------------------------------------------------


        // doublez̍őAŏԂB
        public static void getMinMax(double[] array, ref double min, ref double max)
        {
            int length = array.Length;

            try
            {
                if (length < 1) return;
                min = array[0];
                max = min;
                for (int i = 1; i < length; i++)
                {
                    double value = array[i];
                    if (min > value)
                    {
                        min = value;
                        continue;
                    }
                    if (max < value) max = value;
                }
                return;
            }
            catch (Exception ex)
            {
                MsgBox(ex.ToString());
                return;
            }
        }

        // longz̍őAŏԂB
        public static void getMinMax(long[] array, ref long min, ref long max)
        {
            int length = array.Length;

            try
            {
                if (length < 1) return;
                min = array[0];
                max = min;
                for (int i = 1; i < length; i++)
                {
                    long value = array[i];
                    if (min > value)
                    {
                        min = value;
                        continue;
                    }
                    if (max < value) max = value;
                }
                return;
            }
            catch (Exception ex)
            {
                MsgBox(ex.ToString());
                return;
            }
        }

        // zɈvf}Bl͐ݒ肵ȂB
        //
        // indexŎw肵vf̑Oɑ}B
        //
        public static int insertArray<T>(ref T[] data, int index)
        {
            try
            {
                if (data == null && index == 0)
                {
                    data = new T[1];
                    return 0;
                }

                if (index == data.Length)
                {
                    Array.Resize(ref data, data.Length + 1);
                    return 0;
                }

                if (index > data.Length) return -1;

                Array.Resize(ref data, data.Length + 1);
                Array.Copy(data, index, data, index + 1, data.Length - index - 1);
                return 0;
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return -2;
            }

        }

        //zɈvf}Alݒ肷
        public static int insertArray<T>(ref T[] data, int index, T value)
        {
            try
            {
                int nret = insertArray<T>(ref data, index);
                if (nret < 0) return -1;
                data[index] = value;
                return 0;
            }
            catch (Exception)
            {
                return -1;
            }

        }



        // z񂩂vf폜B
        //
        public static int removeArray<T>(ref T[] data, int index)
        {
            try
            {
                Array.Copy(data, index + 1, data, index, data.Length - index - 1);
                Array.Resize(ref data, data.Length - 1);
                return 0;
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return -1;
            }

        }

        // zŁAs폜B
        //
        public static string[] removeEmptyLines(string[] text)
        {
            try
            {
                string[] lines = new string[ text.Length];
                int n = text.Length;
                int index = 0;
                for(int i=0; i < n;i++ )
                {
                    if ( ! string.IsNullOrEmpty(text[i] ) ) lines[index++]=text[i];
                }
                Array.Resize<string>(ref lines,index);
                return lines;
            }
            catch(Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return null;
            }
        }


        //***************************************************************************
        //                           @@@@lϊ
        //---------------------------------------------------------------------------

        // DoubleɕϊB
        //
        // ϊłȂꍇDouble.MaxValueԂB
        //
        public static double parseDouble(string str)
        {
            double x;
            bool ok = double.TryParse(str, out x);
            if (ok) return x;

            Regex regx = new Regex(@"^[\+\- 0-9.]+[eE]?[0-9.]*$");
            Match m = regx.Match(str);
            if (m != null)
            {
                ok = double.TryParse(m.Value, out x);
                if (ok) return x;
            }

            return double.MaxValue;
        }

        // xxbɕϊ
        //
        public static void deg2dms( double degree, out int deg, out int min, out double sec )
        {
            deg = (int) degree;
            double minsec = (degree - deg) * 60;
            min = (int)minsec;
            sec = (minsec - min) * 60;
        }


        //***************************************************************************
        //                           @@
        //---------------------------------------------------------------------------

        // SppAL𔼊pɂB
        //
        public static string getHankakuString(string str)
        {
            try
            {
                char[] fromChars = str.ToCharArray();
                int numChars = fromChars.Length;

                char[] toChars = new char[numChars];

                for (int i = 0; i < numChars; i++)
                {
                    toChars[i] = getHankakuChar(fromChars[i]);
                }
                return new string(toChars);
            }
            catch (Exception ex)
            {
                MsgBox(ex.ToString());
                return str;
            }
        }

        public static char getHankakuChar(char ch)
        {
            string hankaku = "!\"#$%&'()*+,-./" +
            "0123456789:;<=>?@" +
            "ABCDEFGHIJKLMNOPQRSTUVWXYZ\\]^_`" +
            "abcdefghijklmnopqrstuvwxyz{|}~";

            string zenkaku = "Ihfij{C|D^" +
            "OPQRSTUVWXFGH" +
            "`abcdefghijklmnopqrstuvwxynOQe" +
            "obp`";

            int index = zenkaku.IndexOf(ch);
            if (index >= 0) return hankaku[index];
            else return ch;
        }


        //***************************************************************************
        //                           @@t@C
        //---------------------------------------------------------------------------

        // ۑt@C擾B
        //
        //  fileter:tB^̐Aďc_ (|) ƃtB^ p^[
        //           "text(*.txt) | *.txt | all | *.* "
        public static string getSaveFilePath(string title, string initDirecotory, string filter)
        {
            return getSaveFilePath(title, initDirecotory, "", filter);
        }

        public static string getSaveFilePath(string title, string initDirecotory,
                                        string initFileName, string filter)
        {
            string fileName;
            SaveFileDialog dlg = new SaveFileDialog();

            if (initDirecotory == "")
            {
                if (m_lastAccessDirectory != "") initDirecotory = m_lastAccessDirectory;
                else initDirecotory = "c:\\";
            }
            dlg.InitialDirectory = initDirecotory;
            dlg.FileName = initFileName;
            dlg.Filter = filter;
            dlg.FilterIndex = 1;
            dlg.Title = title;
            dlg.RestoreDirectory = true;

            if (dlg.ShowDialog() != DialogResult.OK) return "";
            fileName = dlg.FileName;
            m_lastAccessDirectory = Path.GetDirectoryName(fileName);
            return fileName;

        }

        // I[vt@C擾B
        public static string getOpenFilePath(string title, string initDirectory, string initFileName,
                                                string filter, bool checkFileExists)
        {
            string fileName;
            OpenFileDialog dlg = new OpenFileDialog();

            if (initDirectory == "")
            {
                if (m_lastAccessDirectory != "") initDirectory = m_lastAccessDirectory;
                else initDirectory = "c:\\";
            }
            dlg.InitialDirectory = initDirectory;
            dlg.FileName = initFileName;
            dlg.Filter = filter;
            dlg.FilterIndex = 1;
            dlg.Title = title;
            dlg.CheckFileExists = checkFileExists;
            dlg.RestoreDirectory = true;

            if (dlg.ShowDialog() != DialogResult.OK) return "";
            fileName = dlg.FileName;
            m_lastAccessDirectory = Path.GetDirectoryName(fileName);
            return fileName;
        }

        // I[vt@C(Ij擾B
        public static string[] getOpenFileNames(string title, string initDirectory, string initFileName,
                                                string filter, bool checkFileExists)
        {
            string[] fileNames;
            OpenFileDialog dlg = new OpenFileDialog();

            if (initDirectory == "")
            {
                if (m_lastAccessDirectory != "") initDirectory = m_lastAccessDirectory;
                else initDirectory = "c:\\";
            }
            dlg.InitialDirectory = initDirectory;
            dlg.FileName = initFileName;
            dlg.Filter = filter;
            dlg.FilterIndex = 1;
            dlg.Title = title;
            dlg.CheckFileExists = checkFileExists;
            dlg.RestoreDirectory = true;
            dlg.Multiselect = true;

            if (dlg.ShowDialog() != DialogResult.OK) return null;
            fileNames = dlg.FileNames;
            if (fileNames.Length > 0) m_lastAccessDirectory = Path.GetDirectoryName(fileNames[0]);
            return fileNames;
        }

        // t@CI[vł邩ׂ
        //
        // ߂l 0:I[v\
        //          -1:t@C݂Ȃ
        //          -2:[hI[̃t@CȂ̂ŏ݂łȂ
        //          -3:ANZX
        //          -4:̗̑RŃI[vłȂ
        //
        public static int canOpen(string filePath, FileAccess readWriteMode, out string errorMessage)
        {
            errorMessage = "";
            try
            {
                if (!File.Exists(filePath))
                {
                    errorMessage = "w肳ꂽt@C܂B";
                    return -1;
                }

                if (readWriteMode != FileAccess.Read)
                {
                    if (File.GetAttributes(filePath) == FileAttributes.ReadOnly)
                    {
                        errorMessage = "w肳ꂽt@C͏݂ł܂B";
                        return -2;
                    }
                }
                FileStream stream = File.Open(filePath, FileMode.Open, readWriteMode);
                if (stream != null)
                {
                    stream.Close();
                    return 0;
                }
            }
            catch (Exception ex)
            {
                errorMessage = ex.Message;
                return -3;
            }

            return 0;
        }

        // gqtBłɂĂꍇ͒u
        public static string addFileExtension(string filePath, string newExtension)
        {
            try
            {
                string body = Path.GetFileNameWithoutExtension(filePath);
                if (newExtension[0] != '.') newExtension = "." + newExtension;
                string newPath = body + newExtension;
                return newPath;
            }
            catch (Exception ex)
            {
                MsgBox(ex.ToString());
                return "";
            }
        }

        //***************************************************************************
        //                          @tH[
        //---------------------------------------------------------------------------

        // tH[\B
        //
        public static void showForm(Form form, int left, int top, Form owner)
        {
            showForm(form, left, top, owner, false);
        }

        public static void showFormDialog(Form form, Form owner, bool dialogMode)
        {
            try
            {
                int left, top;
                if (owner == null)
                {
                    Rectangle rect = SystemInformation.WorkingArea;
                    left = (rect.Width - form.Width) / 2;
                    top = (rect.Height - form.Height) / 2;
                }
                else
                {
                    left = owner.Left + (owner.Width - form.Width) / 2;
                    top = owner.Top + (owner.Height - form.Height) / 2;
                }
                showForm(form, left, top, owner, dialogMode);
            }
            catch (Exception ex)
            {
                MsgBox(ex.ToString());
            }
        }

        public static void showForm(Form form, int left, int top, Form owner, bool dialogMode)
        {
            try
            {
                if (form == null) return;
                form.StartPosition = FormStartPosition.Manual;
                form.Location = new Point(left, top);
                form.Owner = owner;
                if (dialogMode) form.ShowDialog();
                else form.Show();
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
            }
        }

        //***************************************************************************
        //                           @@摜
        //---------------------------------------------------------------------------

        // 摜𑜓xύX
        public static int changeImageSize(string sourcePath, string destPath, int newWidth, int newHeight)
        {
            try
            {
                Bitmap sourceImage =(Bitmap) Image.FromFile(sourcePath);
                Bitmap destImage = new Bitmap(sourceImage, newWidth, newHeight);
                destImage.Save(destPath);

                return 0;
            }
            catch (Exception ex)
            {
                Util.MsgBox(ex.ToString());
                return -1;
            }
        }


        //******************************@vZX@********************************

        //@gqŊ֘AtꂽAvɂA hLg\B
        public static void dispDocument(string fileName, bool waitExit)
        {
            try
            {
                Process process = new Process();
                process.StartInfo.FileName = fileName;
                // process.StartInfo.Verb = "Print";
                process.StartInfo.CreateNoWindow = false;
                process.Start();
                if (waitExit) process.WaitForExit();
            }
            catch (Exception ex)
            {
                MsgBox(ex.ToString());
            }
        }

        //******************************@C^[lbg@********************************

        // F؂Ȃ
        public static string httpRequest(string url)
        {
            return httpRequest(url, "", "");
        }

        public static string httpRequest(string url,Encoding readEncoding)
        {
            return httpRequest(url, "", readEncoding, "", "");
        }

        // BasicF؂URL̓e擾B
        public static string httpRequest(string url, string userName, string password)
        {
            return httpRequest(url, "", userName, password);
        }


        // HTTPŃRec擾B
        //
        // postData : GET̏ꍇ""wB
        //
        public static string httpRequest(string url, string postData, string userName, string password)
        {
            return httpRequest(url, postData, null, userName, password);
        }

        // POST̏ꍇAKvȏꍇURLGR[fBOĂB
        //
        public static string httpRequest(string url, string postData, Encoding readEncoding, string userName, string password)
        {
            return httpRequest(url, postData, readEncoding, userName, password, null);
        }

        public static string httpRequest(string url, string postData,
                    Encoding readEncoding, string userName, string password, CookieContainer cookie)
        {
            return httpRequest(url, postData, readEncoding, userName, password, cookie, "");
        }

        public static string httpRequest(string url, string postData, 
                    Encoding readEncoding, string userName, string password, CookieContainer cookie,
                        string referer)
       {
           try
           {
               HttpWebRequest req = (HttpWebRequest)WebRequest.Create(url);
               req.UserAgent = "Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1)";
               if (cookie != null) req.CookieContainer = cookie;
               if (!string.IsNullOrEmpty(referer)) req.Referer = referer;
               req.Timeout = m_httpTimeoutMilliSec;

               if (!string.IsNullOrEmpty(postData))
               {
                   req.Method = "POST";
                   byte[] byteArray;
                   byteArray = Encoding.UTF8.GetBytes(postData);

                   req.ContentType = "application/x-www-form-urlencoded";
                   req.ContentLength = byteArray.Length;
                   Stream dataStream = req.GetRequestStream();
                   dataStream.Write(byteArray, 0, byteArray.Length);
                   dataStream.Close();
               }

               if (!string.IsNullOrEmpty(userName))
               {
                   // req.PreAuthenticate = true; // ֌WȂ悤

                   //NetworkCredential myCredentials = new NetworkCredential("meidai", "tsurumai","Digest db");
                   NetworkCredential myCredentials = new NetworkCredential(userName, password);
                   req.Credentials = myCredentials;

/*
                   CredentialCache myCache = new CredentialCache();
//                   myCache.Add(new Uri(url), "Digest", new NetworkCredential("meidai", "tsurumai", "Digest db"));
                   myCache.Add(new Uri(url), "Basic", new NetworkCredential("meidai", "tsurumai"));
                   req.Credentials = myCache;
 */
               }

               WebResponse res = req.GetResponse();

               Stream st = res.GetResponseStream();

               StreamReader sr;
               if (readEncoding == null) sr = new StreamReader(st);
               else sr = new StreamReader(st, readEncoding);

               string response = sr.ReadToEnd();
               sr.Close();
               st.Close();

               return response;
           }
           catch (WebException wex)
           {

               string message = wex.Message + "\nServer response = ";
               if (wex.Response == null) message += "nothing";
               else message += wex.Response.ToString();
               message += "\nExcection status = ";
               if (wex.Status == null) message += "nothing";
               else message += wex.Status.ToString();
               MessageBox.Show(message);
               return "";
           }
           catch (Exception ex)
           {
               MessageBox.Show(ex.ToString());
               return "";
           }
        }

        // HTTPANZXł̃^CAEgli~bjݒ肷B
        // 
        //
        public static int httpSetTimeout(int millisec)
        {
            if (millisec >= 1) m_httpTimeoutMilliSec = millisec;
            return m_httpTimeoutMilliSec;
        }


        //***************************************************************************
        //                          @bZ[W\
        //---------------------------------------------------------------------------


        // bZ[W\B
        public static void MsgBox(string message)
        {
            System.Windows.Forms.MessageBox.Show(message);
        }

        // Yes No _CAO\B
        //
        public static bool dialogYes(string message)
        {
            return dialogYes(message, "");
        }



        public static bool dialogYes(string message, string dialogTitle)
        {
            DialogResult yesno = MessageBox.Show(message, dialogTitle, MessageBoxButtons.YesNo);
            return yesno == DialogResult.Yes;
        }

        public static bool dialogNo(string message)
        {
            return dialogNo(message, "");
        }

        public static bool dialogNo(string message, string dialogTitle)
        {
            DialogResult yesno = MessageBox.Show(message, dialogTitle, MessageBoxButtons.YesNo);
            return yesno == DialogResult.No;
        }



        //҂bZ[W\B
        //Rg[ł̕\
        //
        // message : \eLXgB""̏ꍇ͂łɕ\ĂtH[B
        //
        public static void showWaitMessage(string message, Form dispForm)
        {
            try
            {
                if ((m_waitMessageControl != null))
                {
                    m_waitMessageControl.Dispose();
                    if (message == "") return;
                }
                FormWaitMessage f = new FormWaitMessage();

                f.TopLevel = false;
                dispForm.Controls.Add(f);

                f.Top = dispForm.Height / 2 - f.Height / 2;
                f.Left = dispForm.Width / 2 - f.Width / 2;
                f.TextBoxMessage.Text = message;

                f.Show();
                f.BringToFront();
                f.Refresh();
                m_waitMessageControl = f;
                Application.DoEvents();
            }
            catch (Exception ex)
            {
                MsgBox(ex.ToString());
            }
        }

        //҂bZ[W\B
        //tH[ł̕\
        //
        // message : \eLXgB""̏ꍇ͂łɕ\ĂtH[B

        public static void showWaitMessageForm(string message, Form dispForm)
        {
            if (m_waitMessageForm != null)
            {

                if (message == "")
                {
                    m_waitMessageForm.Close();
                    m_waitMessageForm = null;
                    return;
                }
                m_waitMessageForm.TextBoxMessage.Text = message;
                dispForm.Refresh();
                m_waitMessageForm.Refresh();
                Application.DoEvents();
                return;
            }
            else
            {
                FormWaitMessage f = new FormWaitMessage();

                f.TextBoxMessage.Text = message;

                Point pnt = new Point();
                pnt.X = dispForm.Width / 2;
                pnt.Y = dispForm.Height / 2;
                f.Location = pnt;
                f.Owner = dispForm;
                f.Show(dispForm);
                m_waitMessageForm = f;
                f.Refresh();
            }
            return;
        }

        // debug.print
        public static void debugPrint(string message)
        {
            System.Diagnostics.Debug.WriteLine(message);
        }

        /*
        // ͂B
        //
        // ~̎nullԂB
        //
        public static string inputBox(string message, string initValue, Form parentForm)
        {
            try
            {
                FormInputBox formInputBox = new FormInputBox(message, initValue);
                Util.showFormDialog(formInputBox, parentForm, true);
                if (formInputBox.DialogResult == DialogResult.Cancel) return null;
                string inputStr = formInputBox.m_value;
                return inputStr;
            }
            catch (Exception ex)
            {
                MsgBox(ex.ToString());
                return "";
            }
        }
        */


        //***************************************************************************
        //                          @WXg
        //---------------------------------------------------------------------------

        public enum enumRegistry
        {
            CurrentUser, LocalMachine, ClassesRoot, Users, PerfomanceData, CurrentConfig, DynaData
        };


        public static RegistryKey getRootRegistryKey(string rootName)
        {
            try
            {
                string[] rootNames = {"HKEY_CURRENT_USER","HKEY_LOCAL_MACHINE","HKEY_CLASSES_ROOT",
                        "HKEY_USERS","HKEY_PERFORMANCE_DATA","HKEY_CURRENT_CONFIG","HKEY_DYN_DATA" };
                RegistryKey[] rootKeys = { Registry.CurrentUser, Registry.LocalMachine, Registry.ClassesRoot,
                        Registry.Users, Registry.PerformanceData, Registry.CurrentConfig, Registry.DynData };

                rootName = rootName.Trim().ToUpper();
                int index = Array.IndexOf(rootNames, rootName);
                if (index < 0 || index >= rootKeys.Length) return null;
                return rootKeys[index];
            }
            catch (Exception ex)
            {
                MsgBox(ex.ToString());
                return null;
            }
        }

        // w肵WXgpX̃L[擾B
        //
        // pX̍ŏ͎̂ꂩłȂƂȂB
        // "HKEY_CURRENT_USER","HKEY_LOCAL_MACHINE ","HKEY_CLASSES_ROOT ",
        // "HKEY_USERS ","HKEY_PERFORMANCE_DATA ","HKEY_CURRENT_CONFIG ","HKEY_DYN_DATA " 
        //
        // ̊֐Ŏ擾L[̓[hI[B
        // 擾L[͎gpAClose()B
        //
        public static RegistryKey getRegistryKey(string registryPath)
        {
            try
            {
                string[] keyNodeNames = registryPath.Split('\\');
                int numNodes = keyNodeNames.Length;
                if (numNodes < 1) return null;

                RegistryKey rootRegistry = getRootRegistryKey(keyNodeNames[0]);
                if (numNodes == 1) return rootRegistry;

                RegistryKey parentKey = rootRegistry;
                RegistryKey key = null;
                for (int i = 1; i < numNodes; i++)
                {
                    string keyName = keyNodeNames[i];
                    string[] keyNames = parentKey.GetSubKeyNames();
                    if (keyNames.Length < 1) return null;

                    int index = Array.IndexOf(keyNames, keyName);
                    if (index < 0) return null;

                    key = parentKey.OpenSubKey(keyNames[index]);
                    parentKey.Close();
                    if (index == numNodes - 1) break;
                    parentKey = key;
                }
                return key;
            }
            catch (Exception ex)
            {
                MsgBox(ex.ToString());
                return null;
            }
        }

        public static string getRegistryValue(string keyPath, string valueName)
        {
            try
            {
                RegistryKey key = getRegistryKey(keyPath);
                if (key == null) return "";

                Object valueData = key.GetValue(valueName);
                if (valueData == null) return "";
                return valueData.ToString();
            }
            catch (Exception ex)
            {
                MsgBox(ex.ToString());
                return "";
            }
        }


        //***************************************************************************
        //                          @Í
        //---------------------------------------------------------------------------

        //@̈Í̊֐͏URL̃Rs[B
        //http://dobon.net/vb/dotnet/string/encryptstring.html
        // Í

        /// <summary>
        /// Í
        /// </summary>
        /// <param name="str">Í镶</param>
        /// <param name="key">pX[h</param>
        /// <returns>Íꂽ</returns>
        public static string EncryptString(string str, string key)
        {
            //oCg^zɂ
            byte[] bytesIn = System.Text.Encoding.UTF8.GetBytes(str);

            //DESCryptoServiceProviderIuWFNg̍쐬
            System.Security.Cryptography.DESCryptoServiceProvider des =
                new System.Security.Cryptography.DESCryptoServiceProvider();

            //LL[ƏxN^
            //pX[hoCgzɂ
            byte[] bytesKey = System.Text.Encoding.UTF8.GetBytes(key);
            //LL[ƏxN^ݒ
            des.Key = ResizeBytesArray(bytesKey, des.Key.Length);
            des.IV = ResizeBytesArray(bytesKey, des.IV.Length);

            //Íꂽf[^o߂MemoryStream
            System.IO.MemoryStream msOut = new System.IO.MemoryStream();
            //DESÍIuWFNg̍쐬
            System.Security.Cryptography.ICryptoTransform desencrypt =
                des.CreateEncryptor();
            //ނ߂CryptoStream̍쐬
            System.Security.Cryptography.CryptoStream cryptStreem =
                new System.Security.Cryptography.CryptoStream(msOut,
                desencrypt,
                System.Security.Cryptography.CryptoStreamMode.Write);
            //
            cryptStreem.Write(bytesIn, 0, bytesIn.Length);
            cryptStreem.FlushFinalBlock();
            //Íꂽf[^擾
            byte[] bytesOut = msOut.ToArray();

            //
            cryptStreem.Close();
            msOut.Close();

            //Base64ŕɕύXČʂԂ
            return System.Convert.ToBase64String(bytesOut);
        }

        /// <summary>
        /// Íꂽ𕜍
        /// </summary>
        /// <param name="str">Íꂽ</param>
        /// <param name="key">pX[h</param>
        /// <returns>ꂽ</returns>
        public static string DecryptString(string str, string key)
        {
            //DESCryptoServiceProviderIuWFNg̍쐬
            System.Security.Cryptography.DESCryptoServiceProvider des =
                new System.Security.Cryptography.DESCryptoServiceProvider();

            //LL[ƏxN^
            //pX[hoCgzɂ
            byte[] bytesKey = System.Text.Encoding.UTF8.GetBytes(key);
            //LL[ƏxN^ݒ
            des.Key = ResizeBytesArray(bytesKey, des.Key.Length);
            des.IV = ResizeBytesArray(bytesKey, des.IV.Length);

            //Base64ŕoCgzɖ߂
            byte[] bytesIn = System.Convert.FromBase64String(str);
            //Íꂽf[^ǂݍނ߂MemoryStream
            System.IO.MemoryStream msIn =
                new System.IO.MemoryStream(bytesIn);
            //DESIuWFNg̍쐬
            System.Security.Cryptography.ICryptoTransform desdecrypt =
                des.CreateDecryptor();
            //ǂݍނ߂CryptoStream̍쐬
            System.Security.Cryptography.CryptoStream cryptStreem =
                new System.Security.Cryptography.CryptoStream(msIn,
                desdecrypt,
                System.Security.Cryptography.CryptoStreamMode.Read);

            //ꂽf[^擾邽߂StreamReader
            System.IO.StreamReader srOut =
                new System.IO.StreamReader(cryptStreem,
                System.Text.Encoding.UTF8);
            //ꂽf[^擾
            string result = srOut.ReadToEnd();

            //
            srOut.Close();
            cryptStreem.Close();
            msIn.Close();

            return result;
        }

        /// <summary>
        /// LL[pɁAoCgz̃TCYύX
        /// </summary>
        /// <param name="bytes">TCYύXoCgz</param>
        /// <param name="newSize">oCgz̐V傫</param>
        /// <returns>TCYύXꂽoCgz</returns>
        private static byte[] ResizeBytesArray(byte[] bytes, int newSize)
        {
            byte[] newBytes = new byte[newSize];
            if (bytes.Length <= newSize)
            {
                for (int i = 0; i < bytes.Length; i++)
                    newBytes[i] = bytes[i];
            }
            else
            {
                int pos = 0;
                for (int i = 0; i < bytes.Length; i++)
                {
                    newBytes[pos++] ^= bytes[i];
                    if (pos >= newBytes.Length)
                        pos = 0;
                }
            }
            return newBytes;
        }

    }

}