﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using GeoMapLib;
using Newtonsoft.Json.Linq;
using System.IO;
using MathNet.Numerics.LinearAlgebra;
using MathNet.Numerics;
using System.Numerics;

namespace shapeMove
{
    public partial class Form1 : Form
    {
        string yahooId = "ヤフーのAPIキーを指定";
        string googleId = "GoogleのAPIキーを指定";

        double[,] zoneOrigin = new double[19, 2]
        {
            {129.5, 33.0},{131.0, 33.0},{132 + 10.0 / 60, 36.0},
            {133.5, 33.0},{134 + 20.0 / 60, 36.0},{136.0, 36.0},
            {137 + 10.0 / 60, 36.0},{138.5, 36.0},{139 + 50.0 / 60, 36.0},
            {140 + 50.0 / 60, 40.0},{140 + 15.0 / 60, 44.0},{142 + 15.0 / 60, 44.0},
            {144 + 15.0 / 60, 44.0},{142.0, 26.0},{127.5, 26.0},
            {124.0, 26.0},{131.0, 26.0},{136.0, 20.0},
            {154.0, 26.0}
        };

        public Form1()
        {
            InitializeComponent();
        }

         
        private void buttonConvert_Click(object sender, EventArgs e)
        {
            try
            {
                int nret;
                
                string shpPath = Util.getOpenFilePath("", "d:", "", "*.shp|*.shp", true);
                if (shpPath == "") return;

                string shpDir = Path.GetDirectoryName(shpPath);

                string outPath = Util.getSaveFilePath("", shpDir, "*.*|*.*");
                if (outPath == "") return;

                clsShapeFile shapeFile = new clsShapeFile();
                clsShapeFile.clsRecord[] records = new clsShapeFile.clsRecord[0];
                clsShapeFile.stFieldInfo[] dbfFieldInfo;
                int numRecords = clsShapeFile.readAllRecords(shpPath, ref records, false, out dbfFieldInfo);
                if (numRecords < 0)
                {
                    MessageBox.Show("Shape File　リードエラー");
                    return;
                }

                // 投影法
                string projectionInfo;
                nret = clsShapeFile.getProjectionInfo(shpPath, out projectionInfo);
                if (nret < 0)
                {
                    MessageBox.Show("Shape File：Projection リードエラー");
                    return;
                }

                string projStr = "2000_Japan_Zone";
                int pos = projectionInfo.IndexOf(projStr);
                if (pos < 0)
                {
                    MessageBox.Show("Shape File：Zone リードエラー");
                    return;
                }

                int zoneNumber = -1;
                int.TryParse(projectionInfo.Substring(pos + projStr.Length + 1), out zoneNumber);
                if (zoneNumber < 1 || zoneNumber > 19)
                {
                    MessageBox.Show("Shape File：Zone Number リードエラー");
                    return;
                }

                // 市町村名フィールド番号
                int numFields = dbfFieldInfo.Length;
                int cityIndex = -1;
                for (int i = 0; i < numFields; i++)
                {
                    if (dbfFieldInfo[i].name == "市町村名")
                    {
                        cityIndex = i;
                        break;
                    }
                }
                if (cityIndex < 0)
                {
                    MessageBox.Show("Shapeファイルに「市町村名」のカラムがありません。");
                    return;
                }

                // 10個の対応点を取得
                //int numRefPoints = 10;
                int numRefPoints = 20;
                if (numRecords < numRefPoints) numRefPoints = numRecords;
                int delta = numRecords / numRefPoints;
                stXY[] pointsSource = new stXY[numRefPoints];
                stXY[] pointsDest = new stXY[numRefPoints];
                int numValid = 0;
                for (int i = 0; i < numRefPoints; i++)
                {
                    int idx = i * delta;
                    // 中心点
                    clsPoints[] pointsArray = records[idx].points;
                    clsPoints points = pointsArray[0];
                    points.calcCenterMinMax();
                    pointsSource[i] = new stXY(points.centerX, points.centerY);

                    // Geocoding
                    string address = "";
                    for (int j = 0; j < 6; j++)
                    {
                        string column = records[idx].Attributes[cityIndex + j].Trim();
                        if (string.IsNullOrEmpty(column))
                        {
                            if (j == 5) // 地番が無い場合は抜ける
                            {
                                address = "";
                                break;
                            }
                            continue;
                        }
                        address += column;
                    }
                    if (address == "")
                    {
                        pointsDest[i].x = double.NaN;
                        continue;
                    }

                    double yLat, yLon;
                    nret = yahooGeocoder(yahooId, address, out yLat, out yLon);
                    if (nret < 5)
                    {
                        pointsDest[i].x = double.NaN;
                        continue;
                    }

                    double gLat, gLon;
                    nret = googleGeocoder(googleId, address, out gLat, out gLon);
                    if (nret < 5)
                    {
                        pointsDest[i].x = double.NaN;
                        continue;
                    }

                    // ヤフーとGoogleで緯度、経度の差が0.0001（約10m)以下のデータのみ使う
                    double thold = 0.0001;
                    if ( gLat - yLat > thold || gLon - yLon > thold )
                    {
                        pointsDest[i].x = double.NaN;
                        continue;
                    }
                    double lat = (gLat + yLat) / 2.0;
                    double lon = (gLon + yLon) / 2.0;
                    
                    double x, y;
                    nret = latlon2Heimen(lat, lon, zoneNumber, out x, out y);
                    if (nret < 0) continue;
                    pointsDest[i].x = y;
                    pointsDest[i].y = x;
                    numValid++;

                }
                if (numValid < 3)
                {
                    MessageBox.Show("有効な参照点の検索ができません。");
                    return;
                }

               
                //アルゴリズム１
                Matrix<double> src = Matrix<double>.Build.Dense(numValid, 2);
                Matrix<double> dest = Matrix<double>.Build.Dense(numValid, 2);
                int index = 0;
                for (int i = 0; i < numRefPoints; i++)
                {
                        if (double.IsNaN(pointsDest[i].x)) continue;
                    src[index, 0] = pointsSource[i].x;
                    src[index, 1] = pointsSource[i].y;
                    dest[index, 0] = pointsDest[i].x;
                    dest[index, 1] = pointsDest[i].y;
                    index++;
                }
                Matrix<double> coeffi = GetAffineTransformation(src, dest);
                nret = shapeCopy(shpPath, outPath, coeffi);
                if (nret == 0)
                {
                    MessageBox.Show("ファイル変換が正常に終了しました。");
                }
                else
                {
                    MessageBox.Show("ファイル変換に失敗しました");
                }

            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }



        }

        int shapeCopy(string sourceFileName, string destFileName, Matrix<double> matrix)    //アルゴリズム１
        {
            int i, recordNum, nret;
            int fileLength, shapetype, numRecords, recordBytes, numFields;
            double xMin, yMin, xMax, yMax, zMin, zMax, mMin, mMax;
            clsShapeFile.stFieldInfo[] dbfFieldInfo = null;
            DateTime date_;
            clsShapeFile.clsRecord record = null;

            try
            {
                clsShapeFile shape1 = new clsShapeFile();
                clsShapeFile shape2 = new clsShapeFile();

                shape1.open(sourceFileName, false);
                shape2.open(destFileName, true);

                shape1.getShpHeader(out fileLength, out shapetype, out xMin, out yMin,
                         out xMax, out yMax, out zMin, out zMax, out mMin, out mMax);
                shape1.getDbfHeader(out date_, out numRecords, out recordBytes,
                                    out numFields, out dbfFieldInfo);

                shape2.setShpHeader(-1, shapetype, xMin, yMin, xMax, yMax, zMin, zMax, mMin, mMax
                );
                shape2.setShxHeader(-1, shapetype, xMin, yMin, xMax, yMax, zMin, zMax, mMin, mMax
                );
                shape2.setDbfHeader(date_, numRecords, numFields, dbfFieldInfo);
                shape2.write3Headers();

                xMin = double.MaxValue;
                yMin = double.MaxValue;
                xMax = -double.MaxValue;
                yMax = -double.MaxValue;
                numRecords = shape1.getShxNumRecords();
                for (i = 0; i < numRecords; i++)
                {
                    recordNum = shape1.readRecord(out record);
                    if (recordNum < 0) return -1;

                    clsPoints[] pointsArray = record.points;
                    clsPoints points = pointsArray[0];
                    for (int j = 0; j < points.numPoints; j++)
                    {
                        stXY oldPoint = points.points[j];

                        //アルゴリズム１
                        Matrix<double> P = Matrix<double>.Build.DenseOfArray(new double[,]{{ oldPoint.x, oldPoint.y },});
                        Matrix<double> Q = P.Append(Matrix<double>.Build.Dense(P.RowCount, 1, 1));
                        Matrix<double> newPoint = Q * matrix;

                        points.points[j].x = newPoint.At(0, 0);
                        points.points[j].y = newPoint.At(0, 1);
                    }

                    points.calcCenterMinMax();
                    if (xMin > points.minX) xMin = points.minX;
                    if (yMin > points.minY) yMin = points.minY;
                    if (xMax < points.maxX) xMax = points.maxX;
                    if (yMax < points.maxY) yMax = points.maxY;
                    record.xMax = points.maxX;
                    record.xMin = points.minX;
                    record.yMax = points.maxY;
                    record.yMin = points.minY;

                    nret = shape2.writeRecord(recordNum, record);
                    if (nret < 0) return -1;

                }

                shape2.writeDbfEof();
                shape2.setShpHeader(-1, shapetype, xMin, yMin, xMax, yMax, zMin, zMax, mMin, mMax
                );
                shape2.setShxHeader(-1, shapetype, xMin, yMin, xMax, yMax, zMin, zMax, mMin, mMax
                );
                shape2.setDbfHeader(date_, numRecords, numFields, dbfFieldInfo);
                shape2.write3Headers();

                shape1.close();
                shape2.close();

                // 投影法ファイルのコピー
                string sourceProjName = Path.ChangeExtension(sourceFileName, "prj");
                if ( ! File.Exists(sourceProjName))
                {
                    MessageBox.Show("投影法ファイルがありません");
                }
                string destProjName = Path.ChangeExtension(destFileName, "prj");
                File.Copy(sourceProjName, destProjName, true );

                return 0;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
                return -100;
            }
        }

        // Yahoo Map APIを使ってジオコーディングする。
        //
        // 戻り値= マッチングレベル (1:都道府県レベル 2:市区町村レベル 3:町、大字レベル 4:丁目、字レベル
        //           5:詳細レベル1(ex：東京都港区六本木1丁目1)  6:詳細レベル2(ex：東京都港区六本木1丁目1-1)
        //           7:詳細レベル3(ex：東京都港区六本木1丁目1-1-1)
        int yahooGeocoder(string apiKey, string address, out double lat, out double lon)
        {
            int matchingLevel = 0;
            lat = 0;
            lon = 0;
            try
            {
                string url = $"https://map.yahooapis.jp/geocode/V2/geoCoder?appid={apiKey}&query={address}&output=json";


                using (var client = new WebClient())
                {
                    string response = client.DownloadString(url);
                    JObject json = JObject.Parse(response);
                    JArray features = (JArray)json["Feature"];
                    if (features.Count == 0) return 0;
                    int.TryParse((string)features[0]["Property"]["AddressMatchingLevel"], out matchingLevel);
                    string[] lonlatStr = ((string)json["Feature"][0]["Geometry"]["Coordinates"]).Split(',');
                    double.TryParse(lonlatStr[0], out lon);
                    double.TryParse(lonlatStr[1], out lat);

                }
                return matchingLevel;
            }
            catch (Exception ex)
            {
                return -100;
            }
        }

        int googleGeocoder(string apiKey, string address, out double lat, out double lon)
        {
            int matchingLevel = 0;
            lat = 0;
            lon = 0;
            try
            {
                string url = $"https://maps.googleapis.com/maps/api/geocode/json?key={apiKey}&address={address}";


                using (var client = new WebClient())
                {
                    string response = client.DownloadString(url);
                    JObject json = JObject.Parse(response);
                    JArray results = (JArray)json["results"];
                    if (results.Count == 0) return 0;
                    //int.TryParse((string)results[0]["geometry"]["location"], out matchingLevel);
                    string latStr = ((string)results[0]["geometry"]["location"]["lat"]);
                    string lonStr = ((string)results[0]["geometry"]["location"]["lng"]);
                    double.TryParse(lonStr, out lon);
                    double.TryParse(latStr, out lat);
                    matchingLevel = 6;
                }
                return matchingLevel;
            }
            catch (Exception ex)
            {
                return -100;
            }
        }

        // 緯度経度を平面直角座標系に変換する
        // 
        int latlon2Heimen(double lat, double lon, int zoneNumber, out double x, out double y)
        {
            x = 0;
            y = 0;
            try
            {
                if (zoneNumber < 1 || zoneNumber > 19) return -1;
                clsGaussKruger gk = new clsGaussKruger(Const.GRS80);

                double originLat = zoneOrigin[zoneNumber - 1, 1];
                double originLon = zoneOrigin[zoneNumber - 1, 0];
                int nret = gk.latlon2xy(originLat, originLon, 0.9999, lat, lon, out x, out y);
                return nret;
            }
            catch (Exception ex)
            {
                return -100;
            }
        }
 
        Matrix<double> GetAffineTransformation(Matrix<double> src, Matrix<double> dest)
        {
            Matrix<double> A = src.Append(Matrix<double>.Build.Dense(src.RowCount, 1, 1));
            Matrix<double> B = dest.Append(Matrix<double>.Build.Dense(dest.RowCount, 1, 1));

            Matrix<double> X = A.Transpose() * A;
            Matrix<double> Y = A.Transpose() * B;

            Matrix<double> coeffs = X.Inverse() * Y;

            return coeffs;
        }

        private void buttonTest_Click(object sender, EventArgs e)
        {
            //test();
            return;
        }

    }
}
